import torch
torch.set_num_threads(4)

def mpsrg(A, T):

    Asymm = (A + A.permute(2, 1, 0))*0.5
    D, d = Asymm.shape[0], Asymm.shape[1]
    #t0 = time.time()
    B = torch.einsum('ldr,adcb,icj->lairbj', (Asymm, T, Asymm)).contiguous().view(D**2*d, D**2*d)
    C = torch.einsum('ldr,idj->lirj', (Asymm, Asymm)).contiguous().view(D**2, D**2)

    w, _ = torch.symeig(B, eigenvectors=True) 
    lnZ1 = torch.log(w.abs().max())
    w, _ = torch.symeig(C, eigenvectors=True) 
    lnZ2 = torch.log(w.abs().max())

    #lnZ1 = 0.0 
    #lnZ2 = 0.0 
    #for i in range(no_iter):
    #    s = B.norm(1)
    #    lnZ1 = lnZ1 + torch.log(s)/2**i
    #    B = B/s 
    #    B = torch.mm(B, B)

    #    s = C.norm(1)
    #    lnZ2 = lnZ2 + torch.log(s)/2**i
    #    C = C/s
    #    C = torch.mm(C, C)
    return -lnZ1 + lnZ2 

def vmps(T, d, D, no_iter, Nepochs=5):
    
    A = torch.nn.Parameter(0.01*torch.randn(D, d, D, dtype=T.dtype, device=T.device))

    optimizer = torch.optim.LBFGS([A], max_iter=10)
    def closure():
        optimizer.zero_grad()
        #print ('einsum', time.time()- t0)
        #print ((B-B.t()).abs().sum(), (C-C.t()).abs().sum())
        #t0 = time.time()
        loss = mpsrg(A, T.detach()) # loss = -lnZ , here we optimize over A 
        #print ('mpsrg', time.time()- t0)
        print ('            loss', loss.item())
        #t0 = time.time()
        loss.backward(retain_graph=False)
        #print ('backward', time.time()- t0)
        return loss 

    for epoch in range(Nepochs):
        loss = optimizer.step(closure)
        print ('        epoch, free energy', epoch, loss.item())

    return -mpsrg(A.detach(), T), None # pass lnZ out, we need to optimize over T

if __name__=='__main__':
    import time 
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("-D", type=int, default=2, help="D")
    parser.add_argument("-Dcut", type=int, default=20, help="Dcut")
    parser.add_argument("-beta", type=float, default=0.44, help="beta")
    parser.add_argument("-Niter", type=int, default=32, help="Niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="use GPU")
    args = parser.parse_args()
    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    
    #build tensor
    K = torch.tensor([args.beta], dtype=torch.float64, device=device, requires_grad=True)
    c = torch.sqrt(torch.cosh(K))
    s = torch.sqrt(torch.sinh(K))
    M = torch.stack([torch.cat([c, s]), torch.cat([c, -s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M))
    
    #optimization
    lnZ, _ = vmps(T, 2, args.Dcut, args.Niter, args.Nepochs)
    #recompute lnZ using optimized A
    dlnZ = torch.autograd.grad(lnZ, K, create_graph=True)[0] #  En = -d lnZ / d beta
    print (-dlnZ.item())
    #second order derivative evaluated in this way seems not correct, no effect of envioment  
    #dlnZ2 = torch.autograd.grad(dlnZ, K)[0] # Cv = beta^2 * d^2 lnZ / d beta^2 
